import logging
import os.path
import glob
import itertools
import re


def find(base, headers, inheritance='public', many=True,
         required=False):
    '''Scan header files for classes inheriting from a given base class,
    according to the specified inheritance pattern (defaults to 'public').
    Returns a sequence of class names.'''

    regex = '^\s*class\s+([a-zA-Z_]\w+)\s*:\s*' + inheritance + '\s*' + base
    logging.debug('regex: ' + regex)
    cregex = re.compile(regex, re.MULTILINE)

    classes = []
    for header in headers:
        logging.debug('scanning header: ' + header)
        with open(header) as hfile:
            text = hfile.read()
        found_classes = [match.group(1) for match in re.finditer(cregex, text)]
        if found_classes:
            logging.debug(str(len(found_classes)) + ' matches in this file')
        classes += [(c, os.path.realpath(header)) for c in found_classes]

    missing = RuntimeError('could not find any '
                           + base
                           + '-derived class in the specified files.')
    if required and not classes:
        raise missing
    if not many:
        try:
            the_class = classes[0]
        except IndexError:
            if not required:
                return None
            else:
                raise missing
        if len(classes) > 1:
            logging.warning('found several '
                            + base
                            + '-derived classes in the specified files. '
                            'Selecting '
                            + the_class[0] + '(from ' + the_class[1] + ').')
        return the_class
    else:
        return classes


def find_in_dirs(base, dirs, suffixes=[''], inheritance='public', many=True,
                 required=False):
    headers = itertools.chain.from_iterable(
        glob.iglob(os.path.join(directory, '*'+suffix))
        for suffix in suffixes for directory in dirs)
    return find(base, headers,
                inheritance=inheritance, many=many, required=required)


def find_header_for_class(name, dirs, suffixes=['']):
    regex = '^\s*class\s+' + name + '\s*[:{]'
    cregex = re.compile(regex, re.MULTILINE)

    headers = itertools.chain.from_iterable((
        glob.iglob(os.path.join(directory, '*'+suffix))
        for suffix in suffixes for directory in dirs
        ))
    for header in headers:
        with open(header) as hfile:
            text = hfile.read()
        if re.search(cregex, text):
            return header

    return None



#!/usr/bin/env python
# coding: utf-8

import argparse
import sys
import generate_library
import logging
import find_classes

def main():
    parser = argparse.ArgumentParser(
        description='Convert source code for a Geant4 detector construction '
                    'into a geometry that can be read by TRIPOLI-4®.',
        epilog='Any options after `--\' will be passed to the compiler.'
        )

    # general arguments
    g_general = parser.add_argument_group('general arguments')
    g_general.add_argument('-v', '--verbose', help='increase verbosity',
                        action='count', default=0)
    g_general.add_argument('-o', '--output', metavar='OUTPUT_FILE',
                        help='name of the file to generate')
    g_general.add_argument('-I', '--include', metavar='INCLUDE_DIR',
                        help='directories to search for header files (may be '
                                'specified multiple times)',
                        action='append', default=[])
    g_general.add_argument('--geant4-config',
                        help='path to the geant4-config executable')
    g_general.add_argument('sources', metavar='source_file', nargs='*',
                        help='source files')


    # detector args
    g_detector = parser.add_argument_group('detector arguments')
    g_detector.add_argument('-d', '--detector', metavar='DETECTOR_CLASS',
                            help='name of the detector-construction class')
    g_detector.add_argument('--dump-detector-wrapper',
                            help='dump (to stdout) a template for custom wrapper '
                                'functions for the detector class',
                            action='store_true')
    g_detector.add_argument('--custom-detector-wrapper',
                            help='use this if you are providing your custom '
                                'wrapper for detector construction as a source '
                                'file. If this option is not specified, %(prog)s '
                                'will automatically generate a wrapper.',
                            action='store_true')

    # pga args
    g_pga = parser.add_argument_group('primary-generator-action arguments')
    g_pga.add_argument('-p', '--primary-generator-action', metavar='PGA_CLASS',
                    help='name of the primary-generator-action class')
    g_pga.add_argument('--dump-pga-wrapper',
                    help='dump (to stdout) a template for custom wrapper '
                            'functions for the primary generator action class',
                    action='store_true')
    g_pga.add_argument('--custom-pga-wrapper',
                    help='use this if you are providing your custom wrapper '
                            'for the primary generator action as a source file. '
                            'If this option is not specified, %(prog)s will '
                            'automatically generate a wrapper.',
                            action='store_true')

    # parse the arguments
    # limit argument parsing to args before --
    last_arg_to_parse = next((i for i, v in enumerate(sys.argv) if v == '--'),
                            len(sys.argv))
    args = parser.parse_args(sys.argv[1:last_arg_to_parse])
    remaining_args = sys.argv[last_arg_to_parse+1:]

    # dump the detector wrapper template
    if args.dump_detector_wrapper:
        print(generate_library.get_dummy_g42so_detector_wrapper_functions())
        sys.exit(0)

    # dump the wrapper template
    if args.dump_pga_wrapper:
        print(generate_library.get_dummy_g42so_pga_wrapper_functions())
        sys.exit(0)

    # check for the presence of source files
    if not args.sources:
        parser.error('too few arguments')

    # define the logger
    logging.basicConfig(level=logging.WARNING - 10*args.verbose)

    # process the options
    header_suffixes = ['.hh', '.hpp', '.h', '.hxx']

    # process the -d option -- either take the supplied class name or detect it
    if args.detector:
        detector = args.detector
        detector_wheader = (
            detector,
            find_classes.find_header_for_class(
                detector,
                dirs=args.include,
                suffixes=header_suffixes
                )
            )
    else:
        detector_wheader = find_classes.find_in_dirs(
            base='G4VUserDetectorConstruction',
            dirs=args.include,
            suffixes=header_suffixes,
            many=False,
            required=True
            )

    # process the -p option, if supplied
    if args.primary_generator_action:
        pga = args.primary_generator_action
        pga_wheader = (
            pga,
            find_classes.find_header_for_class(
                pga,
                dirs=args.include,
                suffixes=header_suffixes
                )
            )
    else:
        pga_wheader = find_classes.find_in_dirs(
            base='G4VUserPrimaryGeneratorAction',
            dirs=args.include,
            suffixes=header_suffixes,
            many=False,
            required=False)

    generate_library.compile(
        args.sources,
        args.include,
        detector_wheader,
        pga_wheader,
        output=args.output,
        other_flags=remaining_args,
        g4config_path=args.geant4_config,
        custom_detector_wrapper=args.custom_detector_wrapper,
        custom_pga_wrapper=args.custom_pga_wrapper
        )

if __name__ == '__main__':
    main()



from distutils.spawn import find_executable

COMPILERS = ['g++', 'clang++', 'icc']
FLAGS_DICTIONARY = {
    'g++': ['-O2', '-g', '-fPIC', '-shared'],
    'clang++': ['-O2', '-g', '-fPIC', '-shared'],
    'icc': [],  # to be defined
    }


def compiler_and_flags():
    detected = next((comp for comp in COMPILERS if find_executable(comp)),
                    None)
    executable = find_executable(detected)
    return executable, FLAGS_DICTIONARY[detected]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

name = 'g42so'
author = u'Davide Mancusi'
author_email = u'davide.mancusi@cea.fr'
version = '0.1.0'

with open('README.md') as readme:
    long_description = readme.read()

setup(name=name,
      author=author,
      author_email=author_email,
      version=version,
      description='A tool to convert Geant4 geometries into shared libraries',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/arekfu/g42so',
      packages=find_packages(),
      package_data={'': ['detector_wrapper.cc.in',
                         'pga_wrapper.cc.in',
                         'version.cc']},
      python_requires='>=2.6, <3',
      classifiers=[
          "Programming Language :: Python :: 2",
          "Development Status :: 4 - Beta",
          "Environment :: Console",
          "License :: OSI Approved :: MIT License",
          "Operating System :: POSIX :: Linux",
          "Topic :: Scientific/Engineering :: Physics",
          ],
      entry_points={
          'console_scripts': [
              'g42so = g42so.main:main'
              ]
          }
      )



import detect_compiler
from distutils.spawn import find_executable
from pkgutil import get_data
import subprocess
import shlex
import logging
import inspect
import os
import os.path
import sys
import tempfile


def get_g42so_detector_wrapper_functions(d_wh, params=''):
    return get_g42so_wrapper_functions(
        d_wh,
        template_basename='detector_wrapper.cc.in',
        varname='aDetector',
        params=params
    )


def get_g42so_pga_wrapper_functions(pga_wh, params=''):
    return get_g42so_wrapper_functions(
        pga_wh,
        template_basename='pga_wrapper.cc.in',
        varname='aPGA',
        params=params
        )


def get_g42so_wrapper_functions(wh, template_basename, varname='a_var',
                               params='/* parameters go here */'):

    template = get_data('g42so', template_basename)

    include_directives = ['#include "' + wh[1] + '"']
    includes = '\n'.join(include_directives)

    class_name = wh[0]

    wrapper = template.format(includes=includes,
                              varname=varname,
                              class_name=class_name,
                              params=params
                              )

    return wrapper


def get_dummy_g42so_detector_wrapper_functions():
    return get_g42so_detector_wrapper_functions(
        ('MyDetectorConstruction', 'MyDetectorConstruction.hh')
        )


def get_dummy_g42so_pga_wrapper_functions():
    return get_g42so_pga_wrapper_functions(
        ('MyPrimaryGeneratorAction', 'MyPrimaryGeneratorAction.hh')
        )


def write_temp_wrapper_file(wrapper):
    logging.debug('writing wrapper code: ' + wrapper)

    with tempfile.NamedTemporaryFile(suffix='.cc',
                                     mode='w+',
                                     delete=False) as wrapper_file:

        wrapper_file.write(wrapper)
        wrapper_file.flush()
        wrapper_file.seek(0)

        return wrapper_file.name


def compile(sources, includes, d_wh, pga_wh, output=None, other_flags=None,
            g4config_path=None, custom_detector_wrapper=None,
            custom_pga_wrapper=None):
    compiler, flags = detect_compiler.compiler_and_flags()

    # determine the Geant4-specific compilation flags
    if g4config_path:
        g4config = g4config_path
    else:
        g4config = find_executable('geant4-config')
        if not g4config:
            raise RuntimeError('cannot find the geant4-config executable. '
                               'Please specify its location on the command '
                               'line.')
    g4cli = [g4config, '--cflags', '--libs']
    g4process = subprocess.Popen(g4cli, stdout=subprocess.PIPE)
    g4flags_str = g4process.communicate()[0]
    g4flags = shlex.split(g4flags_str)

    # other flags if present
    if not other_flags:
        other_flags = []

    # include dirs
    include_flags = [item for include in includes for item in ['-I', include]]

    # output file
    if not output:
        output = 'lib' + d_wh[0] + '.so'
    logging.info('Will produce the following output file: ' + output)
    output_flags = ['-o', output]

    if not custom_detector_wrapper:
        detector_wrapper = get_g42so_detector_wrapper_functions(d_wh)
        detector_wrapper_file_name = write_temp_wrapper_file(detector_wrapper)
        sources = [detector_wrapper_file_name] + sources
    else:
        detector_wrapper_file_name = None

    if pga_wh and not custom_pga_wrapper:
        pga_wrapper = get_g42so_pga_wrapper_functions(pga_wh)
        pga_wrapper_file_name = write_temp_wrapper_file(pga_wrapper)
        sources = [pga_wrapper_file_name] + sources
    else:
        pga_wrapper_file_name = None

    version_source = get_data('g42so', 'version.cc')
    with tempfile.NamedTemporaryFile(suffix='.cc',
                                     mode='w+') as version_file:
        sources.append(version_file.name)

        # the CLI to execute
        compiler_cli = [compiler] + \
            flags + \
            include_flags + \
            sources + \
            g4flags + \
            other_flags + \
            output_flags

        try:
            logging.info('Running compilation...')
            logging.debug(' ... compiler CLI: ' + ' '.join(compiler_cli))
            subprocess.check_call(compiler_cli)
        finally:
            if detector_wrapper_file_name:
                os.remove(detector_wrapper_file_name)
            if pga_wrapper_file_name:
                os.remove(pga_wrapper_file_name)
